/*
 * Copyright 2013, The University of Waikato.
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 	* Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 	* Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 	* Neither the name of the organization nor the
 *       names of its contributors may be used to endorse or promote products
 * 	  derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * This file implements all the effects
 */

#include "fx.h"
#include "fx-image.h"
#include "fx-effect.h"
#include "clib.h"

#include <ffts.h>
#include <math.h>

#define MIN(a, b) ((a)<(b)?(a):(b))
#define MAX(a, b) ((a)>(b)?(a):(b))

struct fxe_blend_t {
	fxe_t fxe;

	complex float *F;	/* complex filter */

	float sigma;		/* gaussian blur amount */
	float blend;		/* blend amount for srcover rule */
};

/* ********************************************************************** */
/* General purpose kernel types                                           */
/* ********************************************************************** */

typedef struct kernel_t kernel_t;

struct kernel_t {
	// output buffer
	complex float *F;
	/**
	 * Fill one row of complex PSF.  The data must be fft-shifted
	 * ys is the pre-shifted fft Y index
	 */
	void (*fill_row)(kernel_t *k, int ys, complex float *row, int width);

	// for wiener kernel
	float k;
};

/**
 * Create a filter kernel from a PSF
 */
void kernel_set_mt(fxinfo *fx, int id, void *data) {
	kernel_t *k = data;
	int height = fx->fheight;
	int width = fx->fwidth;
	int nthreads = fx->nthreads;
	complex float *row0 = fx->row0 + id * width;

	for (int y = id; y < height; y += nthreads) {
		int ys = y < height / 2 ? y : y - height;

		k->fill_row(k, ys, row0, width);

		ffts_execute(fx->fft_row, row0, fx->tmpa + y * fx->fwidth);
	}

	pthread_barrier_wait(&fx->work_barrier);

	// TODO: mt transpose
	if (id == 0)
		transpose_cf_inplace(fx->tmpa, fx->fwidth, fx->fheight);

	pthread_barrier_wait(&fx->work_barrier);

	for (int x=id;x<fx->fwidth;x+=fx->nthreads) {
		ffts_execute(fx->fft_col, fx->tmpa + x * fx->fheight, k->F + x * fx->fheight);
	}
}

/**
 * Create an approximate wiener deconvolution from a PSF
 */
void wiener_set_mt(fxinfo *fx, int id, void *data) {
	kernel_t *k = data;
	int height = fx->fheight;
	int width = fx->fwidth;
	int nthreads = fx->nthreads;
	complex float *row0 = fx->row0 + id * width;

	for (int y = id; y < height; y += nthreads) {
		int ys = y < height / 2 ? y : y - height;

		k->fill_row(k, ys, row0, width);

		ffts_execute(fx->fft_row, row0, fx->tmpa + y * fx->fwidth);
	}

	pthread_barrier_wait(&fx->work_barrier);

	// TODO: mt transpose
	if (id == 0)
		transpose_cf_inplace(fx->tmpa, fx->fwidth, fx->fheight);

	pthread_barrier_wait(&fx->work_barrier);

	complex float *dst = k->F;
	float K = k->k;
	for (int x=id;x<width;x+=fx->nthreads) {
		int ys = x == 0 ? 1 : 0; // Don't wiener DC!

		ffts_execute(fx->fft_col, fx->tmpa + x * height, dst + x * height);
		for (int y=ys;y<height;y++) {
			complex float v = dst[x * height + y];
			float d = creal(v) * creal(v) + cimag(v) * cimag(v) + K;
			
			dst[x * height + y] = conjf(v) / d;
		}
	}
}

/* ********************************************************************** */
/* Specific PSF functions                                                 */
/* ********************************************************************** */

struct gaussian_kernel_t {
	kernel_t kern;
	float scale;
	float s2;
};

// Standard guassian 2d
static void gaussian_fill_row(kernel_t *k, int ys, complex float *row, int width) {
	struct gaussian_kernel_t *g = (struct gaussian_kernel_t *)k;
	float scale = g->scale;
	float s2 = g->s2;

	for (int x = 0; x < width; x++) {
		int xs = x < width / 2 ? x : x - width;
		float v = scale * expf(s2 * (ys * ys + xs * xs));

		row[x] = v;
	}
}

// Creates a filter which does dx + dy in one pass
// TODO: use it
static void gaussian_dxdy_fill_row(kernel_t *k, int ys, complex float *row, int width) {
	struct gaussian_kernel_t *g = (struct gaussian_kernel_t *)k;
	float scale = g->scale;
	float s2 = g->s2;

	for (int x = 0; x < width; x++) {
		int xs = x < width / 2 ? x : x - width;
		float r2 = (ys * ys + xs * xs);

		row[x] = -xs * scale * expf(-r2 * s2)
			+ I *(-ys * scale * expf(-r2 * s2));
	}
}

// differential of gaussian in x
static void gaussian_dx_fill_row(kernel_t *k, int ys, complex float *row, int width) {
	struct gaussian_kernel_t *g = (struct gaussian_kernel_t *)k;
	float scale = g->scale;
	float s2 = g->s2;

	for (int x = 0; x < width; x++) {
		int xs = x < width / 2 ? x : x - width;
		float r2 = (ys * ys + xs * xs);
		//float v = scale * expf(s2 * (ys * ys + xs * xs));
		float v = -xs * scale * expf(-r2 * s2);

		row[x] = v;
	}
}

// differential of gaussian in y
static void gaussian_dy_fill_row(kernel_t *k, int ys, complex float *row, int width) {
	struct gaussian_kernel_t *g = (struct gaussian_kernel_t *)k;
	float scale = g->scale;
	float s2 = g->s2;

	for (int x = 0; x < width; x++) {
		int xs = x < width / 2 ? x : x - width;
		float r2 = (ys * ys + xs * xs);

		row[x] = -ys * scale * expf(-r2 * s2);
	}
}

/**
 * Create a gaussian kernel
 */
void gaussian_set(fxinfo *fx, float sigma, complex float *F) {
	struct gaussian_kernel_t g = {
		.kern.F = F,
		.kern.fill_row = gaussian_fill_row,
		.scale =  1.0f / (M_PI * 2 * sigma * sigma),
		.s2 = -1.0f / (2 * sigma * sigma)
	};

	LOG("Loading gaussian %f", sigma);
	fx_call_threads(fx, kernel_set_mt, &g);
	LOG("Done loading gaussian %f", sigma);
}

/**
 * Create a pair of differential-of-guassian kernels for each
 * axis direction.
 */
void gaussian_gradient_set(fxinfo *fx, float sigma, complex float *Fdx, complex float *Fdy) {
	struct gaussian_kernel_t gdx = {
		.kern.F = Fdx,
		.kern.fill_row = gaussian_dx_fill_row,
		.scale =  1.0f / (M_PI * 2 * sigma * sigma * sigma * sigma),
		.s2 = 1.0f / (2 * sigma * sigma)
	};
	struct gaussian_kernel_t gdy = {
		.kern.F = Fdy,
		.kern.fill_row = gaussian_dy_fill_row,
		.scale =  1.0f / (M_PI * 2 * sigma * sigma * sigma * sigma),
		.s2 = 1.0f / (2 * sigma * sigma)
	};

	fx_call_threads(fx, kernel_set_mt, &gdx);
	fx_call_threads(fx, kernel_set_mt, &gdy);
}

/**
 * Create a gaussian wiener kernel
 */
void gaussian_wiener_set(fxinfo *fx, float sigma, float k, complex float *F) {
	struct gaussian_kernel_t g = {
		.kern.F = F,
		.kern.fill_row = gaussian_fill_row,
		.kern.k = k,
		.scale =  1.0f / (M_PI * 2 * sigma * sigma),
		.s2 = -1.0f / (2 * sigma * sigma)
	};

	fx_call_threads(fx, wiener_set_mt, &g);
}

/* ********************************************************************** */
/* General purpose fft preview                                            */
/* ********************************************************************** */

/**
 * Sets preview content as fft
 */
static void fft_loadPreview_mt(fxinfo *fx,
			       int id,
			       void *d) {
	const fxe_t * __restrict__ l = d;

	int ox = l->tx - l->overlap/2;
	int oy = l->ty - l->overlap/2;

	int minx = MAX(ox, 0);
	int miny = MAX(oy, 0);
	int maxx = MIN(ox + fx->fwidth, l->width);
	int maxy = MIN(oy + fx->fheight, l->height);

	int sx = MAX(-ox, 0);
	int ex = MIN(fx->fwidth, l->width-ox);
	int sy = MAX(-oy, 0);
	int ey = MIN(fx->fheight, l->height-oy);

	LOG("%2d: load rect %d,%d - %d,%d  x range %d-%d y range %d-%d\n", id, minx, miny, maxx, maxy, sx, ex, sy, ey);

	// FFT of rows for RGBA data
	complex float *row0 = fx->row0 + id * fx->fwidth;
	complex float *row1 = fx->row1 + id * fx->fwidth;

	for (int y=id;y<fx->fheight;y += fx->nthreads) {
		int yi = clampi(oy + y, 0, l->height-1);

		l->raw2cf(l->src + minx * l->spstride + yi * l->sstride, ex - sx, row0 + sx, row1 + sx);

		// Extend start
		if (sx > 0) {
			complex float a = row0[sx];
			complex float b = row1[sx];
			for (int i=0;i<sx;i++) {
				row0[i] = a;
				row1[i] = b;
			}
		}
		// Extend end
		if (ex < fx->fwidth) {
			complex float a = row0[ex-1];
			complex float b = row1[ex-1];
			for (int i=ex-1;i<fx->fwidth;i++) {
				row0[i] = a;
				row1[i] = b;
			}
			
		}

		ffts_execute(fx->fft_row, row0, fx->tmpa + y * fx->fwidth);
		ffts_execute(fx->fft_row, row1, fx->tmpb + y * fx->fwidth);
	}

	// Sync
	pthread_barrier_wait(&fx->work_barrier);

	// FFT columns now

	// TOOD: MT transpose
	if (id == 0)
		transpose_cf_inplace(fx->tmpa, fx->fwidth, fx->fheight);
	else if (id == 1)
		transpose_cf_inplace(fx->tmpb, fx->fwidth, fx->fheight);

	// Sync
	pthread_barrier_wait(&fx->work_barrier);

	for (int x=id;x<fx->fwidth;x+=fx->nthreads)
		ffts_execute(fx->fft_col, fx->tmpa + x * fx->fheight, fx->Ia + x * fx->fheight);

	for (int x=id;x<fx->fwidth;x+=fx->nthreads)
		ffts_execute(fx->fft_col, fx->tmpb + x * fx->fheight, fx->Ib + x * fx->fwidth);
}

static void fft_loadPreview(fxinfo *fx, fxe_t *fxe) {
	fx_call_threads(fx, fft_loadPreview_mt, fxe);
}

/* ********************************************************************** */
/* General purpose convolution base                                       */
/* ********************************************************************** */

/**
 * Implement a basic convolution update
 */
static void convolve_processPreview_mt(fxinfo *fx, int id, void *d) {
	struct fxe_t *a = d;
	complex float *row0 = fx->row0 + id * fx->fheight;
	complex float *row1 = fx->row1 + id * fx->fheight;

	LOG("%2d: apply rect", id);

	// Convolve Ia (first 2 channels)
	for (int x=id;x<fx->fwidth;x+=fx->nthreads) {
		memcpy(row0, fx->Ia + x * fx->fheight, sizeof(complex float) * fx->fheight);
		a->processRow(fx, a, row0, fx->fheight, x, fx->fwidth, FX_RG);
		ffts_execute(fx->ifft_col, row0, fx->tmpa + x * fx->fheight);
	}

	// Convolve Ib (second 2 channels)
	for (int x=id;x<fx->fwidth;x+=fx->nthreads) {
		memcpy(row0, fx->Ib + x * fx->fheight, sizeof(complex float) * fx->fheight);
		a->processRow(fx, a, row0, fx->fheight, x, fx->fwidth, FX_BA);
		ffts_execute(fx->ifft_col, row0, fx->tmpb + x * fx->fheight);
	}

	// Sync
	pthread_barrier_wait(&fx->work_barrier);

	// FIXME: multi-thread
	if (id == 0)
		transpose_cf_inplace(fx->tmpa, fx->fheight, fx->fwidth);
	else if (id == 1)
		transpose_cf_inplace(fx->tmpb, fx->fheight, fx->fwidth);

	// Sync
	pthread_barrier_wait(&fx->work_barrier);

	// IFFT of rows directly to target
	int overlap = a->overlap;
	int wx = MIN(a->width, a->tx+fx->fwidth-overlap/2) - (a->tx);
	float scale = 1.0f / (fx->fwidth * fx->fheight);

	row0 = fx->row0 + id * fx->fwidth;
	row1 = fx->row1 + id * fx->fwidth;

	for (int y=id;y<fx->fheight-overlap;y+=fx->nthreads) {
		if (y + a->ty >= 0 && y + a->ty < a->height) {
			ffts_execute(fx->ifft_row, fx->tmpa + (y+overlap/2) * fx->fwidth, row0);
			ffts_execute(fx->ifft_row, fx->tmpb + (y+overlap/2) * fx->fwidth, row1);

			a->cf2raw(row0 + overlap/2, row1 + overlap/2, wx, a->dst + a->tx * a->dpstride + (a->ty + y)*a->dstride, scale);
		}
	}
}

static void convolve_processPreview(fxinfo *fx, fxe_t *fxe) {
	fx_call_threads(fx, convolve_processPreview_mt, fxe);
}

/**
 * Implement basic convolution tile processing
 *
 * This is the same as calling loadPreview followed by processPreview,
 * but more cache friendly and handles outupt tiling properly.
 */
static void convolve_processTile_mt(fxinfo *fx, int id, void *d) {
	fxe_t * __restrict__ r = d;
	float scale = 1.0f / (fx->fwidth * fx->fheight);

	int ox = r->tx - r->overlap/2;
	int oy = r->ty - r->overlap/2;

	int minx = MAX(ox, 0);
	int miny = MAX(oy, 0);
	int maxx = MIN(ox + fx->fwidth, r->width);
	int maxy = MIN(oy + fx->fheight, r->height);

	int sx = MAX(-ox, 0);
	int ex = MIN(fx->fwidth, r->width-ox);
	int sy = MAX(-oy, 0);
	int ey = MIN(fx->fheight, r->height-oy);

	LOG("convolve rect %d,%d - %d,%d  x range %d-%d y range %d-%d\n", minx, miny, maxx, maxy, sx, ex, sy, ey);

	complex float *row0 = fx->row0 + id * fx->fwidth;
	complex float *row1 = fx->row1 + id * fx->fwidth;

	// FFT of rows for RGBA data
	for (int y=id;y<fx->fheight;y+=fx->nthreads) {
		int yi = clampi(oy + y, 0, r->height-1);

		r->raw2cf(r->src + minx * r->spstride + yi * r->sstride, ex - sx, row0 + sx, row1 + sx);

		// Extend start
		if (sx > 0) {
			complex float a = row0[sx];
			complex float b = row1[sx];
			for (int i=0;i<sx;i++) {
				row0[i] = a;
				row1[i] = b;
			}
		}
		// Extend end
		if (ex < fx->fwidth) {
			complex float a = row0[ex-1];
			complex float b = row1[ex-1];
			for (int i=ex-1;i<fx->fwidth;i++) {
				row0[i] = a;
				row1[i] = b;
			}
			
		}

		ffts_execute(fx->fft_row, row0, fx->tmpa + y * fx->fwidth);
		ffts_execute(fx->fft_row, row1, fx->tmpb + y * fx->fwidth);
	}

	pthread_barrier_wait(&fx->work_barrier);

	if (id == 0)
		transpose_cf_inplace(fx->tmpa, fx->fwidth, fx->fheight);
	else if (id == 1)
		transpose_cf_inplace(fx->tmpb, fx->fwidth, fx->fheight);

	pthread_barrier_wait(&fx->work_barrier);

	row0 = fx->row0 + id * fx->fheight;
	row1 = fx->row1 + id * fx->fheight;

	// Convolve tmpa (first 2 channels)
	for (int x=id;x<fx->fwidth;x+=fx->nthreads) {
		ffts_execute(fx->fft_col, fx->tmpa + x * fx->fheight, row0);
		r->processRow(fx, r, row0, fx->fheight, x, fx->fwidth, FX_RG);
		ffts_execute(fx->ifft_col, row0, fx->tmpa + x * fx->fheight);
	}
	// Convolve tmpb (second 2 channels)
	for (int x=id;x<fx->fwidth;x+=fx->nthreads) {
		ffts_execute(fx->fft_col, fx->tmpb + x * fx->fheight, row0);
		r->processRow(fx, r, row0, fx->fheight, x, fx->fwidth, FX_BA);
		ffts_execute(fx->ifft_col, row0, fx->tmpb + x * fx->fheight);
	}

	pthread_barrier_wait(&fx->work_barrier);

	if (id == 0)
		transpose_cf_inplace(fx->tmpa, fx->fheight, fx->fwidth);
	else if (id == 1)
		transpose_cf_inplace(fx->tmpb, fx->fheight, fx->fwidth);

	pthread_barrier_wait(&fx->work_barrier);

	// IFFT of rows
	int overlap = r->overlap;
	int width = r->width;
	int height = r->height;
	int tx = r->tx;
	int ty = r->ty;
	int wx = MIN(width, tx+fx->fwidth-overlap/2) - (tx);

	row0 = fx->row0 + id * fx->fwidth;
	row1 = fx->row1 + id * fx->fwidth;

	for (int y=id;y<fx->fheight-overlap;y+=fx->nthreads) {
		if (y + ty >= 0 && y + ty < height) {
			ffts_execute(fx->ifft_row, fx->tmpa + (y+overlap/2) * fx->fwidth, row0);
			ffts_execute(fx->ifft_row, fx->tmpb + (y+overlap/2) * fx->fwidth, row1);

			r->cf2raw(row0 + overlap/2, row1 + overlap/2, wx, r->dst + tx * r->dpstride + (ty + y)*r->dstride, scale);
		}
	}
}

static void convolve_processTile(fxinfo *fx, fxe_t *fxe) {
	fx_call_threads(fx, convolve_processTile_mt, fxe);
}

static void convolve_processRow(fxinfo *fx, fxe_t *fxe, complex float *row, int len, int xi, int xlen, fx_channel_t cid) {
	struct fxe_blend_t *d = (struct fxe_blend_t *)fxe;
	complex float *F = d->F + xi * xlen;

	for (int i=0;i<len;i++)
		row[i] = row[i] * F[i];
}

/* ********************************************************************** */
/* Blur & blend                                                           */
/* ********************************************************************** */

static void blend_update(fxinfo *fx, fxe_t *fxe, float *data, int datalen, float *params, int paramlen) {
	struct fxe_blend_t *b = (struct fxe_blend_t *)fxe;
	int changed = 0;

	float sigma = params[0];
	float blend = params[1];

	changed = (sigma != b->sigma);

	b->sigma = sigma;
	b->blend = blend;

	if (changed)
		gaussian_set(fx, sigma, b->F);
}

static void blend_processRow(fxinfo *fx, fxe_t *fxe, complex float *row, int len, int xi, int xlen, fx_channel_t cid) {
	struct fxe_blend_t *d = (struct fxe_blend_t *)fxe;
	complex float *F = d->F + xi * xlen;
	float NA = d->blend;
	float A = 1.0f - NA;

	for (int i=0;i<len;i++)
		row[i] = (A * row[i] + NA * row[i] * F[i]);
}

static void blend_free(fxinfo *fx, fxe_t *fxe) {
	struct fxe_blend_t *b = (struct fxe_blend_t *)fxe;

	free(b->F);

	free(fxe);
}

/**
 * Create a blended/blur effect.
 */
fxe_t *fxe_blend_init(fxinfo *fx, float *data, int datalen, float *params, int paramlen) {
	struct fxe_blend_t *blend = malloc(sizeof(*blend));

	blend->fxe.fx = fx;
	blend->fxe.update = blend_update;
	blend->fxe.loadPreview = fft_loadPreview;
	blend->fxe.processPreview = convolve_processPreview;
	blend->fxe.processTile = convolve_processTile;
	blend->fxe.processRow = blend_processRow;
	blend->fxe.free = blend_free;

	blend->F = xmemalign(64, fx->fwidth * fx->fheight * sizeof(*blend->F));

	blend_update(fx, &blend->fxe, data, datalen, params, paramlen);

	return &blend->fxe;
}

/* ********************************************************************** */
/* Wiener deconvoltion                                                    */
/* ********************************************************************** */

struct fxe_wiener_t {
	fxe_t fxe;

	complex float *F;	/* complex filter */

	float sigma;		/* gaussian blur amount */
	float k;		/* noise suppression factor */
	//float blend;		/* blend amount for srcover rule (would this be useful?) */
};


static void wiener_update(fxinfo *fx, fxe_t *fxe, float *data, int datalen, float *params, int paramlen) {
	struct fxe_wiener_t *b = (struct fxe_wiener_t *)fxe;
	int changed = 0;

	float sigma = params[0];
	float k = params[1];

	changed = (sigma != b->sigma) | (k != b->k);

	b->sigma = sigma;
	b->k = k;

	if (changed)
		gaussian_wiener_set(fx, sigma, k, b->F);
}

static void wiener_free(fxinfo *fx, fxe_t *fxe) {
	struct fxe_wiener_t *b = (struct fxe_wiener_t *)fxe;

	free(b->F);

	free(fxe);
}

/**
 * Gaussian wiener deconvolution
 */
fxe_t *fxe_wiener_init(fxinfo *fx, float *data, int datalen, float *params, int paramlen) {
	struct fxe_wiener_t *fxe = malloc(sizeof(*fxe));

	fxe->fxe.fx = fx;
	fxe->fxe.update = wiener_update;
	fxe->fxe.loadPreview = fft_loadPreview;
	fxe->fxe.processPreview = convolve_processPreview;
	fxe->fxe.processTile = convolve_processTile;
	fxe->fxe.processRow = convolve_processRow;
	fxe->fxe.free = wiener_free;

	fxe->F = xmemalign(64, fx->fwidth * fx->fheight * sizeof(*fxe->F));

	wiener_update(fx, &fxe->fxe, data, datalen, params, paramlen);

	return &fxe->fxe;
}

/* ********************************************************************** */
/* Gradient based stuff                                                   */
/* ********************************************************************** */

struct fxe_gradient_t {
	struct fxe_t fxe;

	complex float *Fdx;	/* x differential filter */
	complex float *Fdy;	/* y differential filter */

	void (*processOutput)(fxinfo *fx, struct fxe_gradient_t *fxe, complex float *dxdy, int len, int y, int stride, complex float *outrg, complex float *outba);
};

/**
 * Loads preview for gradient functions.
 *
 * This creates an FFT of a greyscale version of the input
 */
static void gradient_loadPreview_mt(fxinfo *fx,
				    int id,
				    void *d) {
	const fxe_t * __restrict__ l = d;

	int ox = l->tx - l->overlap/2;
	int oy = l->ty - l->overlap/2;

	int minx = MAX(ox, 0);
	int miny = MAX(oy, 0);
	int maxx = MIN(ox + fx->fwidth, l->width);
	int maxy = MIN(oy + fx->fheight, l->height);

	int sx = MAX(-ox, 0);
	int ex = MIN(fx->fwidth, l->width-ox);
	int sy = MAX(-oy, 0);
	int ey = MIN(fx->fheight, l->height-oy);

	LOG("%2d: load rect %d,%d - %d,%d  x range %d-%d y range %d-%d\n", id, minx, miny, maxx, maxy, sx, ex, sy, ey);

	complex float *row0 = fx->row0 + id * fx->fwidth;
	complex float *row1 = fx->row1 + id * fx->fwidth;
	complex float *row2 = fx->row2 + id * fx->fwidth;

	for (int y=id;y<fx->fheight;y += fx->nthreads) {
		int yi = clampi(oy + y, 0, l->height-1);

		l->raw2cf(l->src + minx * l->spstride + yi * l->sstride, ex - sx, row0 + sx, row1 + sx);

		// Extend start
		if (sx > 0) {
			complex float a = row0[sx];
			complex float b = row1[sx];
			for (int i=0;i<sx;i++) {
				row0[i] = a;
				row1[i] = b;
			}
		}
		// Extend end
		if (ex < fx->fwidth) {
			complex float a = row0[ex-1];
			complex float b = row1[ex-1];
			for (int i=ex-1;i<fx->fwidth;i++) {
				row0[i] = a;
				row1[i] = b;
			}			
		}

		// Save copy of float signal
		memcpy(fx->ia + fx->fwidth * y, row0, fx->fwidth * sizeof(*fx->ia));
		memcpy(fx->ib + fx->fwidth * y, row1, fx->fwidth * sizeof(*fx->ib));

		// Convert to greyscale
		for (int x=0;x<fx->fwidth;x++) {
			float r = creal(row0[x]);
			float g = cimag(row0[x]);
			float b = creal(row1[x]);

			row0[x] = r * 0.299f + g * 0.587f + b * 0.114f;
		}

		ffts_execute(fx->fft_row, row0, fx->tmpa + y * fx->fwidth);
	}

	// Sync
	pthread_barrier_wait(&fx->work_barrier);

	// FFT columns now

	// TOOD: MT transpose
	if (id == 0)
		transpose_cf_inplace(fx->tmpa, fx->fwidth, fx->fheight);

	// Sync
	pthread_barrier_wait(&fx->work_barrier);

	for (int x=id;x<fx->fwidth;x+=fx->nthreads)
		ffts_execute(fx->fft_col, fx->tmpa + x * fx->fheight, fx->Ia + x * fx->fheight);
}

static void gradient_loadPreview(fxinfo *fx, fxe_t *fxe) {
	fx_call_threads(fx, gradient_loadPreview_mt, fxe);
}

/**
 * Filter the greyscale input with the 2 gradient filters
 */
static void gradient_processPreview_mt(fxinfo *fx, int id, void *d) {
	struct fxe_gradient_t *a = d;
	complex float *row0 = fx->row0 + id * fx->fheight;
	complex float *row1 = fx->row1 + id * fx->fheight;
	complex float *row2 = fx->row2 + id * fx->fheight;
	complex float *Fdx = a->Fdx;
	complex float *Fdy = a->Fdy;

	LOG("%2d: apply rect", id);

	// Convolve against two filters
	for (int x=id;x<fx->fwidth;x+=fx->nthreads) {
		memcpy(row2, fx->Ia + x * fx->fheight, sizeof(complex float) * fx->fheight);
		// convolve with each pair of filters
		for (int y=0;y<fx->fheight;y++) {
			row0[y] = row2[y] * Fdx[x * fx->fheight + y];
			row1[y] = row2[y] * Fdy[x * fx->fheight + y];
		}

		ffts_execute(fx->ifft_col, row0, fx->tmpa + x * fx->fheight);
		ffts_execute(fx->ifft_col, row1, fx->tmpb + x * fx->fheight);
	}

	// Sync
	pthread_barrier_wait(&fx->work_barrier);

	// FIXME: multi-thread
	if (id == 0)
		transpose_cf_inplace(fx->tmpa, fx->fheight, fx->fwidth);
	else if (id == 1)
		transpose_cf_inplace(fx->tmpb, fx->fheight, fx->fwidth);

	// Sync
	pthread_barrier_wait(&fx->work_barrier);

	// IFFT of rows directly to target
	int overlap = a->fxe.overlap;
	int wx = MIN(a->fxe.width, a->fxe.tx+fx->fwidth-overlap/2) - (a->fxe.tx);
	float scale = 1.0f / (fx->fwidth * fx->fheight);

	row0 = fx->row0 + id * fx->fwidth;
	row1 = fx->row1 + id * fx->fwidth;

	for (int y=id;y<fx->fheight-overlap;y+=fx->nthreads) {
		if (y + a->fxe.ty >= 0 && y + a->fxe.ty < a->fxe.height) {
			ffts_execute(fx->ifft_row, fx->tmpa + (y+overlap/2) * fx->fwidth, row0);
			ffts_execute(fx->ifft_row, fx->tmpb + (y+overlap/2) * fx->fwidth, row1);

			for (int x=0;x<fx->fwidth;x++)
				row2[x] = (creal(row0[x]) + I * creal(row1[x])) * scale;

			a->processOutput(fx, a, row2, fx->fwidth, y, fx->fwidth, row0, row1);

			a->fxe.cf2raw(row0 + overlap/2, row1 + overlap/2, wx, a->fxe.dst + a->fxe.tx * a->fxe.dpstride + (a->fxe.ty + y)*a->fxe.dstride, 1.0f);
		}
	}
}

static void gradient_processPreview(fxinfo *fx, fxe_t *fxe) {
	fx_call_threads(fx, gradient_processPreview_mt, fxe);
}

static void gradient_processTile_mt(fxinfo *fx,
				    int id,
				    void *d) {
	const fxe_t * __restrict__ l = d;

	int ox = l->tx - l->overlap/2;
	int oy = l->ty - l->overlap/2;

	int minx = MAX(ox, 0);
	int miny = MAX(oy, 0);
	int maxx = MIN(ox + fx->fwidth, l->width);
	int maxy = MIN(oy + fx->fheight, l->height);

	int sx = MAX(-ox, 0);
	int ex = MIN(fx->fwidth, l->width-ox);
	int sy = MAX(-oy, 0);
	int ey = MIN(fx->fheight, l->height-oy);

	LOG("%2d: load rect %d,%d - %d,%d  x range %d-%d y range %d-%d\n", id, minx, miny, maxx, maxy, sx, ex, sy, ey);

	complex float *row0 = fx->row0 + id * fx->fwidth;
	complex float *row1 = fx->row1 + id * fx->fwidth;
	complex float *row2 = fx->row2 + id * fx->fwidth;

	for (int y=id;y<fx->fheight;y += fx->nthreads) {
		int yi = clampi(oy + y, 0, l->height-1);

		l->raw2cf(l->src + minx * l->spstride + yi * l->sstride, ex - sx, row0 + sx, row1 + sx);

		// Extend start
		if (sx > 0) {
			complex float a = row0[sx];
			complex float b = row1[sx];
			for (int i=0;i<sx;i++) {
				row0[i] = a;
				row1[i] = b;
			}
		}
		// Extend end
		if (ex < fx->fwidth) {
			complex float a = row0[ex-1];
			complex float b = row1[ex-1];
			for (int i=ex-1;i<fx->fwidth;i++) {
				row0[i] = a;
				row1[i] = b;
			}			
		}

		// Save copy of float signal
		memcpy(fx->ia + fx->fwidth * y, row0, fx->fwidth * sizeof(*fx->ia));
		memcpy(fx->ib + fx->fwidth * y, row1, fx->fwidth * sizeof(*fx->ib));

		// Convert to greyscale
		for (int x=0;x<fx->fwidth;x++) {
			float r = creal(row0[x]);
			float g = cimag(row0[x]);
			float b = creal(row1[x]);

			row0[x] = r * 0.299f + g * 0.587f + b * 0.114f;
		}

		ffts_execute(fx->fft_row, row0, fx->tmpa + y * fx->fwidth);
	}

	// Sync
	pthread_barrier_wait(&fx->work_barrier);

	// FFT columns now

	// TOOD: MT transpose
	if (id == 0)
		transpose_cf_inplace(fx->tmpa, fx->fwidth, fx->fheight);

	// Sync
	pthread_barrier_wait(&fx->work_barrier);

	//for (int x=id;x<fx->fwidth;x+=fx->nthreads)
	//	ffts_execute(fx->fft_col, fx->tmpa + x * fx->fheight, fx->Ia + x * fx->fheight);


	struct fxe_gradient_t *a = d;
	complex float *Fdx = a->Fdx;
	complex float *Fdy = a->Fdy;

	// Convolve against two filters
	for (int x=id;x<fx->fwidth;x+=fx->nthreads) {
		ffts_execute(fx->fft_col, fx->tmpa + x * fx->fheight, row2);
		//memcpy(row2, fx->Ia + x * fx->fheight, sizeof(complex float) * fx->fheight);
		// convolve with each pair of filters
		for (int y=0;y<fx->fheight;y++) {
			row0[y] = row2[y] * Fdx[x * fx->fheight + y];
			row1[y] = row2[y] * Fdy[x * fx->fheight + y];
		}

		ffts_execute(fx->ifft_col, row0, fx->tmpa + x * fx->fheight);
		ffts_execute(fx->ifft_col, row1, fx->tmpb + x * fx->fheight);
	}

	// Sync
	pthread_barrier_wait(&fx->work_barrier);

	// FIXME: multi-thread
	if (id == 0)
		transpose_cf_inplace(fx->tmpa, fx->fheight, fx->fwidth);
	else if (id == 1)
		transpose_cf_inplace(fx->tmpb, fx->fheight, fx->fwidth);

	// Sync
	pthread_barrier_wait(&fx->work_barrier);

	// IFFT of rows directly to target
	int overlap = a->fxe.overlap;
	int wx = MIN(a->fxe.width, a->fxe.tx+fx->fwidth-overlap/2) - (a->fxe.tx);
	float scale = 1.0f / (fx->fwidth * fx->fheight);

	row0 = fx->row0 + id * fx->fwidth;
	row1 = fx->row1 + id * fx->fwidth;

	for (int y=id;y<fx->fheight-overlap;y+=fx->nthreads) {
		if (y + a->fxe.ty >= 0 && y + a->fxe.ty < a->fxe.height) {
			ffts_execute(fx->ifft_row, fx->tmpa + (y+overlap/2) * fx->fwidth, row0);
			ffts_execute(fx->ifft_row, fx->tmpb + (y+overlap/2) * fx->fwidth, row1);

			for (int x=0;x<fx->fwidth;x++)
				row2[x] = (creal(row0[x]) + I * creal(row1[x])) * scale;

			a->processOutput(fx, a, row2, fx->fwidth, y + overlap/2, fx->fwidth, row0, row1);

			a->fxe.cf2raw(row0 + overlap/2, row1 + overlap/2, wx, a->fxe.dst + a->fxe.tx * a->fxe.dpstride + (a->fxe.ty + y)*a->fxe.dstride, 1.0f);
		}
	}
}

static void gradient_processTile(fxinfo *fx, fxe_t *fxe) {
	fx_call_threads(fx, gradient_processTile_mt, fxe);
}

/* ********************************************************************** */
/* Glare effect                                                           */
/* ********************************************************************** */

struct fxe_glare_t {
	struct fxe_gradient_t fxe;

	float sigma;		/* gaussian blur amount */
	float Ir, Ig, Ib;	/* highlight colour */
	float amount;		/* effect strength */
};

static void glare_update(fxinfo *fx, fxe_t *fxe, float *data, int datalen, float *params, int paramlen) {
	struct fxe_glare_t *b = (struct fxe_glare_t *)fxe;
	int changed = 0;

	float sigma = params[0];
	float Ir = params[1];
	float Ib = params[2];
	float Ig = params[3];
	float amount = params[4];

	changed = (sigma != b->sigma);

	b->sigma = sigma;
	b->Ir = Ir * 255.0f;
	b->Ig = Ig * 255.0f;
	b->Ib = Ib * 255.0f;
	// Ia?
	b->amount = amount * 0.2f;

	if (changed) {
		gaussian_gradient_set(fx, sigma, b->fxe.Fdx, b->fxe.Fdy);
	}
}

static void glare_processOutput(fxinfo *fx, struct fxe_gradient_t *fxe, complex float *dxdy, int len, int y, int stride, complex float *outrg, complex float *outba) {
	struct fxe_glare_t *glare = (struct fxe_glare_t *)fxe;
	complex float Irg = (glare->Ir + I * glare->Ig) * glare->amount;
	complex float Iba = (glare->Ib + I * 0) * glare->amount;
	complex float *ia = fx->ia + y * stride;
	complex float *ib = fx->ib + y * stride;

	// Glare processing
	for (int x=0;x<len;x++) {
		float mag = cabsf(dxdy[x]);

		outrg[x] = ia[x] + Irg * mag;
		outba[x] = ib[x] + Iba * mag;
	}
}

static void glare_free(fxinfo *fx, fxe_t *fxe) {
	struct fxe_glare_t *b = (struct fxe_glare_t *)fxe;

	free(b->fxe.Fdx);
	free(b->fxe.Fdy);

	free(fxe);
}

/**
 * Glare effect
 */
fxe_t *fxe_glare_init(fxinfo *fx, float *data, int datalen, float *params, int paramlen) {
	struct fxe_glare_t *fxe = malloc(sizeof(*fxe));
	struct fxe_t *f = &fxe->fxe.fxe;

	f->fx = fx;
	f->update = glare_update;
	f->loadPreview = gradient_loadPreview;
	f->processPreview = gradient_processPreview;
	f->processTile = gradient_processTile;
	//fxe->fxe.processRow = glare_processRow;
	fxe->fxe.processOutput = glare_processOutput;
	f->free = glare_free;

	fxe->fxe.Fdx = xmemalign(64, fx->fwidth * fx->fheight * sizeof(complex float));
	fxe->fxe.Fdy = xmemalign(64, fx->fwidth * fx->fheight * sizeof(complex float));

	glare_update(fx, f, data, datalen, params, paramlen);

	return f;
}

/* ********************************************************************** */
/* Glazed effect                                                          */
/* ********************************************************************** */

typedef struct vec3f vec3f;
struct vec3f {
	float x, y, z;
};

struct fxe_glaze_t {
	struct fxe_gradient_t fxe;

	float sigma;		/* gaussian blur amount */
	float ka, kd, ks;	/* ambient, diffuse, specular */
	float shininess;
	float Dr, Dg, Db;	/* diffuse highlight colour */
	float Sr, Sg, Sb;	/* specular highlight colour */
	vec3f light;
	vec3f viewer;
};

float vec_length(const vec3f * __restrict__ v) {
	return sqrtf(v->x * v->x + v->y * v->y + v->z * v->z);
}

void vec_norm(vec3f * __restrict__ v) {
	float norm = 1.0f / vec_length(v);

	v->x *= norm;
	v->y *= norm;
	v->z *= norm;
}

float vec_dot(const vec3f * __restrict__ a, const vec3f * __restrict__ b) {
	return a->x * b->x + a->y * b->y + a->z * b->z;
}

static void glaze_update(fxinfo *fx, fxe_t *fxe, float *data, int datalen, float *params, int paramlen) {
	struct fxe_glaze_t *b = (struct fxe_glaze_t *)fxe;
	int changed = 0;

	float sigma = params[0];

	changed = (sigma != b->sigma);

	b->sigma = sigma;

	b->ka = params[1];
	b->kd = params[2];
	b->ks = params[3];
	b->shininess = params[4];
	b->Dr = params[5] * 255;
	b->Dg = params[6] * 255;
	b->Db = params[7] * 255;
	b->Sr = params[8] * 255;
	b->Sg = params[9] * 255;
	b->Sb = params[10] * 255;
	b->light.x = params[11];
	b->light.y = params[12];
	b->light.z = params[13];
	b->viewer.x = params[14];
	b->viewer.y = params[15];
	b->viewer.z = params[16];

	vec_norm(&b->light);
	vec_norm(&b->viewer);

	LOG("Glaze config: ka=%f kd=%f ks=%f a=%f Dr=%f Dg=%f Db=%f Sr=%f Sg=%f Sb=%f",
	    b->ka, b->kd, b->ks, b->shininess, b->Dr, b->Dg, b->Db, b->Sr, b->Sg, b->Sb);
	LOG("Glaze config: light=%f %f %f viewer=%f %f %f",
	    b->light.x, b->light.y, b->light.z,
	    b->viewer.x, b->viewer.y, b->viewer.z);

	if (changed) {
		gaussian_gradient_set(fx, sigma, b->fxe.Fdx, b->fxe.Fdy);
	}
}

static void glaze_processOutput(fxinfo *fx, struct fxe_gradient_t * __restrict__ fxe, complex float *dxdy, int len, int y, int stride, complex float *outrg, complex float *outba) {
	struct fxe_glaze_t * __restrict__ glare = (struct fxe_glaze_t *)fxe;
	complex float *ia = fx->ia + y * stride;
	complex float *ib = fx->ib + y * stride;
	float ka = glare->ka;
	float kd = glare->kd;
	float ks = glare->ks;
	float shininess = glare->shininess;

	for (int x=0;x<len;x++) {
		vec3f n;
		vec3f r;

		// Synthesise normalised normal vector from image gradients
		n.x = creal(dxdy[x]);
		n.y = cimag(dxdy[x]);
		n.z = -1;
		vec_norm(&n);

		float ldn = vec_dot(&glare->light, &n);

		// Reflection vector
		r.x = n.x * ldn * 2 - glare->light.x;
		r.y = n.y * ldn * 2 - glare->light.y;
		r.z = n.z * ldn * 2 - glare->light.z;

		float rdv = fabs(vec_dot(&glare->viewer, &r));

		float R = creal(ia[x]);
		float G = cimag(ia[x]);
		float B = creal(ib[x]);
		float A = cimag(ib[x]);

		// Phong shading model
#if 0
		float Ir = ka * R + kd * ldn * glare->Dr + ks * powf(rdv, shininess) * glare->Sr;
		float Ig = ka * G + kd * ldn * glare->Dg + ks * powf(rdv, shininess) * glare->Sg;
		float Ib = ka * B + kd * ldn * glare->Db + ks * powf(rdv, shininess) * glare->Sb;
#else
		float Ir = ka * R + kd * ldn * R + ks * powf(rdv, shininess) * glare->Sr;
		float Ig = ka * G + kd * ldn * G + ks * powf(rdv, shininess) * glare->Sg;
		float Ib = ka * B + kd * ldn * B + ks * powf(rdv, shininess) * glare->Sb;
#endif
		outrg[x] = Ir + Ig * I;
		outba[x] = Ib + A * I;
	}
}

static void glaze_free(fxinfo *fx, fxe_t *fxe) {
	struct fxe_glaze_t *b = (struct fxe_glaze_t *)fxe;

	free(b->fxe.Fdx);
	free(b->fxe.Fdy);

	free(fxe);
}

/**
 * Glaze effect
 */
fxe_t *fxe_glaze_init(fxinfo *fx, float *data, int datalen, float *params, int paramlen) {
	struct fxe_glaze_t *fxe = malloc(sizeof(*fxe));
	struct fxe_t *f = &fxe->fxe.fxe;

	f->fx = fx;
	f->update = glaze_update;
	f->loadPreview = gradient_loadPreview;
	f->processPreview = gradient_processPreview;
	f->processTile = gradient_processTile;
	fxe->fxe.processOutput = glaze_processOutput;
	f->free = glaze_free;

	fxe->fxe.Fdx = xmemalign(64, fx->fwidth * fx->fheight * sizeof(complex float));
	fxe->fxe.Fdy = xmemalign(64, fx->fwidth * fx->fheight * sizeof(complex float));

	glaze_update(fx, f, data, datalen, params, paramlen);

	return f;
}

/* ********************************************************************** */
/* Public interfaces                                                      */
/* ********************************************************************** */

/**
 * Free effect data
 */
void fxe_free(fxinfo *fx, fxe_t *fxe) {
	fxe->free(fx, fxe);
}

/**
 * Update parameters
 */
void fxe_update(fxinfo *fx, fxe_t *fxe, float *data, int datalen, float *params, int paramlen) {
	fxe->update(fx, fxe, data, datalen, params, paramlen);
}

/**
 * Load a preview tile
 */
void fxe_loadPreview(fxinfo *fx, fxe_t *fxe, void *src, int spstride, int sstride, int width, int height, int tx, int ty, int overlap, fraw_to_cf_t raw2cf) {
	fxe->src = src;
	fxe->spstride = spstride;
	fxe->sstride = sstride;
	fxe->width = width;
	fxe->height = height;
	fxe->tx = tx;
	fxe->ty = ty;
	fxe->overlap = overlap;
	fxe->raw2cf = raw2cf;

	fxe->loadPreview(fx, fxe);
}

/**
 * Generate a preview output
 */
void fxe_processPreview(fxinfo *fx, fxe_t *fxe, void *dst, int dpstride, int dstride, int width, int height, int tx, int ty, int overlap, fcf_to_raw_t cf2raw) {
	fxe->dst = dst;
	fxe->dpstride = dpstride;
	fxe->dstride = dstride;
	fxe->width = width;
	fxe->height = height;
	fxe->tx = tx;
	fxe->ty = ty;
	fxe->overlap = overlap;
	fxe->cf2raw = cf2raw;

	fxe->processPreview(fx, fxe);
}

/**
 * Process one tile of overlapped data
 */
void fxe_processTile(fxinfo *fx, fxe_t *fxe,
		     void *src, int spstride, int sstride,
		     void *dst, int dpstride, int dstride,
		     int width, int height, int tx, int ty, int overlap,
		     fraw_to_cf_t raw2cf,
		     fcf_to_raw_t cf2raw) {
	fxe->src = src;
	fxe->spstride = spstride;
	fxe->sstride = sstride;

	fxe->dst = dst;
	fxe->dpstride = dpstride;
	fxe->dstride = dstride;

	fxe->width = width;
	fxe->height = height;
	fxe->tx = tx;
	fxe->ty = ty;
	fxe->overlap = overlap;

	fxe->raw2cf = raw2cf;
	fxe->cf2raw = cf2raw;

	fxe->processTile(fx, fxe);
}
