/*
 * Copyright 2013, The University of Waikato.
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 	* Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 	* Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 	* Neither the name of the organization nor the
 *       names of its contributors may be used to endorse or promote products
 * 	  derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _FX_EFFECT_H
#define _FX_EFFECT_H

#include "clib.h"

/**
 * Data for blend operation
 */
typedef struct ofxe_blend_t {
	// Blend amount, 0-1
	float blend;
} ofxe_blend_t;

/**
 * Hi/low pass pyramid filter parameters
 */
typedef struct fxe_filter_pyramid_t {
	// Line equation slope
	float m;
	// Line equation offset
	float c;
	// Scaling factor for max frequency
	float fix;
} fxe_filter_pyramid_t;

/**
 * Gradient based effects
 */
typedef struct fxe_gradeient_t {
} fxe_gradeient_t;

/**
 * Basic convolution
 */
void fxe_convolve(fxinfo *fx, complex float *row, int len, int xi, int xlen, fx_channel_t cid, void *data);
void fxe_convolve_blend(fxinfo *fx, complex float *row, int len, int xi, int xlen, fx_channel_t cid, void *data);


/**
 * High/low pass filtering
 */
void fxe_filter_pyramid_init(fxe_filter_pyramid_t *f, float low, float high, int samples);
void fxe_filter_pyramid(fxinfo *fx, complex float *row, int len, int xi, int xlen, fx_channel_t cid, void *data);


/**
 * Effect classes
 */
typedef struct fxe_effect_t fxe_effect_t;
typedef struct fxe_t fxe_t;

typedef void (*fxe_process_row_t)(fxinfo *fx, complex float *row, int len, int xi, int xlen, fx_channel_t cid, void *data);

/**
 * Data for effects, effects manage these/subclass them
 */
struct fxe_t {
	fxinfo *fx;

	/**
	 * Set the effect parameters.
	 * May update any pre-calculated data, e.g. fft of filter, etc.
	 */
	void (*update)(fxinfo *fx, fxe_t *fxe, float *data, int datalen, float *params, int paramlen);

	/**
	 * Set preview content.
	 */
	void (*loadPreview)(fxinfo *fx, fxe_t *fxe);
	/**
	 * Calculate preview output.
	 */
	void (*processPreview)(fxinfo *fx, fxe_t *fxe);
	/**
	 * Process a tile of output.
	 */
	void (*processTile)(fxinfo *fx, fxe_t *fxe);
	/**
	 * Process a single row of data.
	 */
	void (*processRow)(fxinfo *fx, fxe_t *fxe, complex float *row, int len, int xi, int xlen, fx_channel_t cid);

	/**
	 * Free fxe.
	 */
	void (*free)(fxinfo *fx, fxe_t *fxe);

	void *src;		/* src pointer */
	int spstride;		/* src element stride in bytes */
	int sstride;		/* src row stride in bytes */
	
	void *dst;		/* dst pointer */
	int dpstride;
	int dstride;

	int width;		/* size of roi */
	int height;
	int tx;			/* location of roi relative to src */
	int ty;

	int overlap;		/* overlap if needed */

	fraw_to_cf_t raw2cf;	/* input data to complex-float */
	fcf_to_raw_t cf2raw;	/* complex-float to output data */
};

void fxe_free(fxinfo *fx, fxe_t *fxe);

void fxe_update(fxinfo *fx, fxe_t *fxe, float *data, int datalen, float *params, int paramlen);
void fxe_loadPreview(fxinfo *fx, fxe_t *fxe, void *src, int spstride, int sstride, int width, int height, int tx, int ty, int overlap, fraw_to_cf_t raw2cf);
void fxe_processPreview(fxinfo *fx, fxe_t *fxe, void *dst, int dpstride, int dstride, int width, int height, int tx, int ty, int overlap, fcf_to_raw_t cf2raw);
void fxe_processTile(fxinfo *fx, fxe_t *fxe,
		     void *src, int spstride, int sstride,
		     void *dst, int dpstride, int dstride,
		     int width, int height, int tx, int ty, int overlap,
		     fraw_to_cf_t raw2cf,
		     fcf_to_raw_t cf2raw);

fxe_t *fxe_blend_init(fxinfo *fx, float *data, int datalen, float *params, int paramlen);
fxe_t *fxe_wiener_init(fxinfo *fx, float *data, int datalen, float *params, int paramlen);
fxe_t *fxe_glare_init(fxinfo *fx, float *data, int datalen, float *params, int paramlen);

#endif
