/*
 * Copyright 2013, The University of Waikato.
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 	* Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 	* Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 	* Neither the name of the organization nor the
 *       names of its contributors may be used to endorse or promote products
 * 	  derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <android/bitmap.h>

#include <math.h>

#include "nz_ac_waikato_picfx_fx_EffectEngine.h"

#include "fx.h"
#include "clib.h"
#include "fx-image.h"
#include "fx-effect.h"

static void throwException(JNIEnv *env, const char *type, const char *what) {
	jclass jc;

	LOG("throwing exception %s `%s'\n", type, what);

	jc = (*env)->FindClass(env, type);
	if (jc)
		(*env)->ThrowNew(env, jc, what);
	else
		LOG("Can't find class `%s'\n", type, what);
}

JNIEXPORT jlong JNICALL Java_nz_ac_waikato_picfx_fx_EffectEngine_create
(JNIEnv *env, jclass jc, jint fwidth, jint fheight, jint nthreads) {
	fxinfo *fx;

	fx = fx_create(fwidth, fheight, nthreads);

	LOG("create %dx%d = %p  row0=%p row=%p", fwidth, fheight, fx, fx->row0, fx->row1);

	return (jlong)(int)fx;
}

JNIEXPORT void JNICALL Java_nz_ac_waikato_picfx_fx_EffectEngine_free
(JNIEnv *env, jclass jc, jlong jp) {
	fxinfo *fx = (fxinfo *)jp;

	if (fx)
		fx_free(fx);
}

JNIEXPORT void JNICALL Java_nz_ac_waikato_picfx_fx_EffectEngine_setProcess
(JNIEnv *env, jclass jc, jlong jp, jint effect, jfloatArray jdata, jfloatArray jparams) {
	fxinfo *fx = (fxinfo *)jp;
	int datalen = 0;
	int paramlen = 0;
	jfloat *data = NULL;
	jfloat params[32];

	if (jdata) {
		datalen = (*env)->GetArrayLength(env, jdata);
		data = (*env)->GetPrimitiveArrayCritical(env, jdata, NULL);
	}
	if (jparams) {
		paramlen = (*env)->GetArrayLength(env, jparams);
		if (paramlen > 32)
			paramlen = 32;
		(*env)->GetFloatArrayRegion(env, jparams, 0, paramlen, params);
	}	

	if (fx->efx_type == effect) {
		fxe_update(fx, fx->efx, data, datalen, params, paramlen);
	} else {
		if (fx->efx) {
			fxe_free(fx, fx->efx);
			fx->efx = NULL;
		}

		fx->efx_type = effect;

		switch (effect) {
		case nz_ac_waikato_picfx_fx_EffectEngine_EFFECT_BLUR:
			fx->efx = fxe_blend_init(fx, data, datalen, params, paramlen);
			break;
		case nz_ac_waikato_picfx_fx_EffectEngine_EFFECT_WIENER:
			fx->efx = fxe_wiener_init(fx, data, datalen, params, paramlen);
			break;
		case nz_ac_waikato_picfx_fx_EffectEngine_EFFECT_GLARE:
			fx->efx = fxe_glare_init(fx, data, datalen, params, paramlen);
			break;
		case nz_ac_waikato_picfx_fx_EffectEngine_EFFECT_GLAZE:
			fx->efx = fxe_glaze_init(fx, data, datalen, params, paramlen);
			break;
		default:
			fx->efx_type = -1;
			throwException(env, "java/lang/UnsupportedOperationException", "Unsupported effect id");
			break;
		}
	}

	if (data)
		(*env)->ReleasePrimitiveArrayCritical(env, jdata, data, 0);
}

JNIEXPORT void JNICALL Java_nz_ac_waikato_picfx_fx_EffectEngine_process
(JNIEnv *env, jclass jc, jlong jp, jobject jsrc, jobject jdst, jint overlap) {
	fxinfo *fx = (fxinfo *)jp;
	AndroidBitmapInfo sinfo;
	AndroidBitmapInfo dinfo;
	uint8_t *spixels;
	uint8_t *dpixels;
	// Data conversions + operation
	fraw_to_cf_t raw2cf;
	fcf_to_raw_t cf2raw;
	int spstride, dpstride;

	AndroidBitmap_getInfo(env, jsrc, &sinfo);
	AndroidBitmap_getInfo(env, jdst, &dinfo);

	LOG("convolve src %dx%d fmt %2d stride %d", sinfo.width, sinfo.height, sinfo.format, sinfo.stride);
	LOG("         dst %dx%d fmt %2d stride %d", dinfo.width, dinfo.height, dinfo.format, dinfo.stride);

	if (sinfo.width != dinfo.width
	    || sinfo.height != dinfo.height) {
		throwException(env, "java/lang/UnsupportedOperationException", "Image size mismatch.");
		return;
	}

	switch (sinfo.format) {
	case ANDROID_BITMAP_FORMAT_RGBA_8888:
		raw2cf = (fraw_to_cf_t)int8888_to_cf;
		spstride = 4;
		break;
	case ANDROID_BITMAP_FORMAT_RGB_565:
		raw2cf = (fraw_to_cf_t)short565_to_cf;
		spstride = 2;
		break;
	default:
		throwException(env, "java/lang/UnsupportedOperationException", "Unsupported source format");
		return;
	}

	switch (dinfo.format) {
	case ANDROID_BITMAP_FORMAT_RGBA_8888:
		cf2raw = (fcf_to_raw_t)cf_to_int8888;
		dpstride = 4;
		break;
	case ANDROID_BITMAP_FORMAT_RGB_565:
		cf2raw = (fcf_to_raw_t)cf_to_short565;
		dpstride = 2;
		break;
	default:
		throwException(env, "java/lang/UnsupportedOperationException", "Unsupported destination format");
		return;
	}

	if (!AndroidBitmap_lockPixels(env, jsrc, (void **)&spixels)) {
		if (!AndroidBitmap_lockPixels(env, jdst, (void **)&dpixels)) {

			//fx_convolve_rgba(fx, spixels, sinfo.stride, dpixels, dinfo.stride, sinfo.width, sinfo.height);
			//memcpy(dpixels, spixels, 100 * sinfo.stride);
			for (int ty=0;ty<dinfo.height;ty+=fx->fheight-overlap) {
				for (int tx=0;tx<dinfo.width;tx+=fx->fwidth-overlap) {
					LOG("process %d,%d", tx, ty);
					fxe_processTile(fx, fx->efx,
							spixels, spstride, sinfo.stride,
							dpixels, dpstride, dinfo.stride,
							dinfo.width, dinfo.height, tx, ty, overlap,
							raw2cf, cf2raw);
				}
			}
			
			AndroidBitmap_unlockPixels(env, jdst);
		} else {
			// error
			throwException(env, "java/lang/RuntimeException", "Unable to lock destination bitmap.");
		}
		AndroidBitmap_unlockPixels(env, jsrc);
	} else {
		// can't lock
		throwException(env, "java/lang/RuntimeException", "Unable to lock source bitmap.");
	}
}

JNIEXPORT void JNICALL Java_nz_ac_waikato_picfx_fx_EffectEngine_update
(JNIEnv *env, jclass jc, jlong jp, jfloatArray jdata, jfloatArray jparams) {
	fxinfo *fx = (fxinfo *)jp;
	int datalen = 0;
	int paramlen = 0;
	jfloat *data = NULL;
	jfloat params[32];

	if (!fx->efx) {
		throwException(env, "java/lang/UnsupportedOperationException", "Effect mode not set");
		return;
	}

	if (jdata) {
		datalen = (*env)->GetArrayLength(env, jdata);
		data = (*env)->GetPrimitiveArrayCritical(env, jdata, NULL);
	}
	if (jparams) {
		paramlen = (*env)->GetArrayLength(env, jparams);
		if (paramlen > 32)
			paramlen = 32;
		(*env)->GetFloatArrayRegion(env, jparams, 0, paramlen, params);
	}

	fxe_update(fx, fx->efx, data, datalen, params, paramlen);

	if (data)
		(*env)->ReleasePrimitiveArrayCritical(env, jdata, data, 0);
}

JNIEXPORT void JNICALL Java_nz_ac_waikato_picfx_fx_EffectEngine_loadPreview
(JNIEnv *env, jclass jc, jlong jp, jobject jsrc, jint tx, jint ty) {
	fxinfo *fx = (fxinfo *)jp;
	AndroidBitmapInfo sinfo;
	uint8_t *spixels;
	fraw_to_cf_t raw2cf;
	int spstride;

	if (!fx->efx) {
		throwException(env, "java/lang/UnsupportedOperationException", "Effect mode not set");
		return;
	}

	AndroidBitmap_getInfo(env, jsrc, &sinfo);

	LOG("set preview src %dx%d fmt %2d stride %d", sinfo.width, sinfo.height, sinfo.format, sinfo.stride);

	switch (sinfo.format) {
	case ANDROID_BITMAP_FORMAT_RGBA_8888:
		raw2cf = (fraw_to_cf_t)int8888_to_cf;
		spstride = 4;
		break;
	case ANDROID_BITMAP_FORMAT_RGB_565:
		raw2cf = (fraw_to_cf_t)short565_to_cf;
		spstride = 2;
		break;
	default:
		throwException(env, "java/lang/UnsupportedOperationException", "Unsupported source format");
		return;
	}

	if (!AndroidBitmap_lockPixels(env, jsrc, (void **)&spixels)) {
		fxe_loadPreview(fx, fx->efx, spixels, spstride, sinfo.stride, sinfo.width, sinfo.height, tx, ty, 0, raw2cf);
		
		AndroidBitmap_unlockPixels(env, jsrc);
	} else {
		// error
		throwException(env, "java/lang/RuntimeException", "Unable to lock destination bitmap.");
	}
}

JNIEXPORT void JNICALL Java_nz_ac_waikato_picfx_fx_EffectEngine_processPreview
(JNIEnv *env, jclass jc, jlong jp, jobject jdst) {
	fxinfo *fx = (fxinfo *)jp;
	AndroidBitmapInfo dinfo;
	uint8_t *dpixels;
	fcf_to_raw_t cf2raw;
	int dpstride;

	if (!fx->efx) {
		throwException(env, "java/lang/UnsupportedOperationException", "Effect mode not set");
		return;
	}

	AndroidBitmap_getInfo(env, jdst, &dinfo);

	LOG("apply    dst %dx%d fmt %2d stride %d", dinfo.width, dinfo.height, dinfo.format, dinfo.stride);

	switch (dinfo.format) {
	case ANDROID_BITMAP_FORMAT_RGBA_8888:
		cf2raw = (fcf_to_raw_t)cf_to_int8888;
		dpstride = 4;
		break;
	case ANDROID_BITMAP_FORMAT_RGB_565:
		cf2raw = (fcf_to_raw_t)cf_to_short565;
		dpstride = 2;
		break;
	default:
		throwException(env, "java/lang/UnsupportedOperationException", "Unsupported destination format");
		return;
	}

	if (!AndroidBitmap_lockPixels(env, jdst, (void **)&dpixels)) {
		fxe_processPreview(fx, fx->efx, dpixels, dpstride, dinfo.stride,
				   dinfo.width, dinfo.height, 0, 0, 0,
				   cf2raw);
		AndroidBitmap_unlockPixels(env, jdst);
	} else {
		// can't lock
		throwException(env, "java/lang/RuntimeException", "Unable to lock target bitmap.");
	}
}
