/*
 * Copyright 2013, The University of Waikato.
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 	* Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 	* Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 	* Neither the name of the organization nor the
 *       names of its contributors may be used to endorse or promote products
 * 	  derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* Low level image fragments */

#include "fx-image.h"

//#define ALIGNED __attribute__((aligned(64)))
#define ALIGNED

void cmult(complex float * ALIGNED dst,
	   complex float * ALIGNED a,
	   complex float * ALIGNED b,
	   int len) {
	for (int i=0;i<len;i++)
		dst[i] = a[i] * b[i];
}

void int8888_to_cf(uint8_t *a,
		int len,
		complex float * ALIGNED dsta,
		complex float * ALIGNED dstb) {
	for (int i=0;i<len;i++) {
		dsta[i] = a[i*4+0] + a[i*4+1]*I;
		dstb[i] = a[i*4+2] + a[i*4+3]*I;
	}
}

void short565_to_cf(uint16_t *src, int len, complex float *dsta, complex float *dstb) {
	// This comes out brighter than it should

	for (int i=0;i<len;i++) {
		uint16_t v = src[i];
		unsigned int r = (v>>11) & 0x1f;
		unsigned int g = (v>>5) & 0x3f;
		unsigned int b = (v) & 0x1f;

		dsta[i] = r * (255.0f / 31.0f) + g * (255.0f / 63.0f) * I;
		dstb[i] = b * (255.0f / 31.0f) + 0 * I;
	}
}

void cf_to_int8888(complex float * ALIGNED srca,
		complex float * ALIGNED srcb,
		int len,
		uint8_t *dst,
		float scale) {
	for (int i=0;i<len;i++) {
		complex float a = srca[i] * scale;
		complex float b = srcb[i] * scale;
		dst[i*4+0] = clampf(creal(a), 0, 255);
		dst[i*4+1] = clampf(cimag(a), 0, 255);
		dst[i*4+2] = clampf(creal(b), 0, 255);
		//dst[i*4+3] = clampf(cimag(b), 0, 255);
		dst[i*4+3] = 255;
	}
}

void cf_to_short565(complex float *srca, complex float *srcb, int len, uint16_t *dst, float scale) {
	for (int i=0;i<len;i++) {
		complex float A = srca[i] * scale;
		complex float B = srcb[i] * scale;
		int r = clampf(creal(A), 0, 255.0f);
		int g = clampf(cimag(A), 0, 255.0f);
		int b = clampf(creal(B), 0, 255.0f);

		r &= 0xf8;
		g &= 0xfc;
		b &= 0xf8;

		dst[i] = (r<<8) | (g<<3) | (b>>3);
	}
}

/**
 * Tranpsoses in-place.
 * Saves memory
 */
void transpose_cf_inplace(complex float * ALIGNED src,
			  int w,
			  int h) {
	// 8x8 is 64x8 bytes and should match cache line
	complex float tmpa[8*8]  __attribute__((aligned(64)));
	complex float tmpb[8*8]  __attribute__((aligned(64)));

	for (int ty=0;ty<h;ty+=8) {
		for (int tx=ty;tx<w;tx+=8) {
			// tx,ty
			for (int y=0;y<8;y++) {
				for (int x=0;x<8;x++) {
					tmpa[y+x*8] = src[tx+x+(ty+y)*w];
				}
			}
			// ty,tx
			if (tx != ty) {
				for (int y=0;y<8;y++) {
					for (int x=0;x<8;x++) {
						tmpb[y+x*8] = src[ty+x+(tx+y)*w];
					}
				}
				for (int y=0;y<8;y++) {
					memcpy(src+tx+(ty+y)*w, tmpb+y*8, 8*sizeof(tmpb[0]));
				}
			}
			for (int y=0;y<8;y++) {
				memcpy(src+ty+(tx+y)*w, tmpa+y*8, 8*sizeof(tmpa[0]));
			}
		}
	}
}

void transpose_rgba_inplace(void *srcp,
			 int w,
			 int h) {
	unsigned int *src = srcp;
	// 8x8 is 64x8 bytes and should match cache line
	unsigned int tmpa[8*8]  __attribute__((aligned(64)));
	unsigned int tmpb[8*8]  __attribute__((aligned(64)));

	for (int ty=0;ty<h;ty+=8) {
		for (int tx=ty;tx<w;tx+=8) {
			// tx,ty
			for (int y=0;y<8;y++) {
				for (int x=0;x<8;x++) {
					tmpa[y+x*8] = src[tx+x+(ty+y)*w];
				}
			}
			// ty,tx
			if (tx != ty) {
				for (int y=0;y<8;y++) {
					for (int x=0;x<8;x++) {
						tmpb[y+x*8] = src[ty+x+(tx+y)*w];
					}
				}
				for (int y=0;y<8;y++) {
					memcpy(src+tx+(ty+y)*w, tmpb+y*8, 8*sizeof(tmpb[0]));
				}
			}
			for (int y=0;y<8;y++) {
				memcpy(src+ty+(tx+y)*w, tmpa+y*8, 8*sizeof(tmpa[0]));
			}
		}
	}
}
