/*
 * Copyright 2013, The University of Waikato.
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 	* Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 	* Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 	* Neither the name of the organization nor the
 *       names of its contributors may be used to endorse or promote products
 * 	  derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "fx.h"
#include "clib.h"

#include "fx-image.h"
#include "fx-effect.h"

#include <ffts.h>

#ifdef ENABLE_THREADS
#include <pthread.h>

struct fx_thread_init {
	int id;
	fxinfo *fx;
};

static void *fx_thread(void *);
#endif

#define MIN(a, b) ((a)<(b)?(a):(b))
#define MAX(a, b) ((a)>(b)?(a):(b))

// TODO: FFT cache?

fxinfo *fx_create(int fwidth, int fheight, int nthreads) {
	fxinfo *fx = malloc(sizeof(*fx));

	LOG("Create FFT FX %dx%d\n", fwidth, fheight);

#ifdef ENABLE_THREADS
	// the non-threaded transpose code needs 2 threads at the moment
	if (nthreads < 2)
		nthreads = 2;
#endif

	fx->fwidth = fwidth;
	fx->fheight = fheight;

	fx->row0 = xmemalign(64, MAX(fwidth, fheight)*sizeof(*fx->row0) * nthreads);
	fx->row1 = xmemalign(64, MAX(fwidth, fheight)*sizeof(*fx->row1) * nthreads);
	fx->row2 = xmemalign(64, MAX(fwidth, fheight)*sizeof(*fx->row2) * nthreads);

	fx->tmpa = xmemalign(64, fwidth*fheight*sizeof(*fx->tmpa));
	fx->tmpb = xmemalign(64, fwidth*fheight*sizeof(*fx->tmpb));

	fx->Ia = xmemalign(64, fwidth*fheight*sizeof(*fx->Ia));
	fx->Ib = xmemalign(64, fwidth*fheight*sizeof(*fx->Ib));

	fx->ia = xmemalign(64, fwidth*fheight*sizeof(*fx->ia));
	fx->ib = xmemalign(64, fwidth*fheight*sizeof(*fx->ib));

	fx->filter = xmemalign(64, fwidth*fheight*sizeof(*fx->filter));
	fx->filter_enable = 0;

	fx->fft_row = ffts_init_1d(fwidth, -1);
	fx->ifft_row = ffts_init_1d(fwidth, 1);

	fx->fft_col = ffts_init_1d(fheight, -1);
	fx->ifft_col = ffts_init_1d(fheight, 1);

	fx->efx_type = -1;
	fx->efx = NULL;

#ifdef ENABLE_THREADS
	fx->nthreads = nthreads;
	pthread_barrier_init(&fx->cmd_barrier, NULL, nthreads+1);
	pthread_barrier_init(&fx->work_barrier, NULL, nthreads);
	for (int i=0;i<nthreads;i++) {
		struct fx_thread_init *ti = malloc(sizeof(*ti));

		ti->id = i;
		ti->fx = fx;
		pthread_create(&fx->threads[i], NULL, fx_thread, ti);
	}
#endif
	return fx;
}

void fx_free(fxinfo *fx) {
#ifdef ENABLE_THREADS
	fx->cmd = FX_QUIT;
	pthread_barrier_wait(&fx->cmd_barrier);
	for (int i=0;i<fx->nthreads;i++) {
		pthread_join(fx->threads[i], NULL);
	}
	pthread_barrier_destroy(&fx->cmd_barrier);
	pthread_barrier_destroy(&fx->work_barrier);
#endif

	if (fx->efx)
		fxe_free(fx, fx->efx);

	ffts_free(fx->fft_row);
	ffts_free(fx->ifft_row);
	ffts_free(fx->fft_col);
	ffts_free(fx->ifft_col);

	free(fx->filter);
	free(fx->ib);
	free(fx->ia);
	free(fx->Ib);
	free(fx->Ia);
	free(fx->tmpb);
	free(fx->tmpa);
	free(fx->row2);
	free(fx->row1);
	free(fx->row0);

	free(fx);
}

#ifdef ENABLE_THREADS
static void *fx_thread(void *d) {
	struct fx_thread_init *f = d;
	fxinfo *fx = f->fx;

	while (1) {
		// Wait for broadcast command
		pthread_barrier_wait(&fx->cmd_barrier);

		switch (fx->cmd) {
		case FX_QUIT:
			goto done;
		case FX_CALL:
			LOG("Thread %d processing", f->id);
			fx->cmd_call(fx, f->id, fx->cmd_data);
			LOG("Thread %d complete", f->id);
			break;
		}

		// Sync with main thread
		pthread_barrier_wait(&fx->cmd_barrier);
	}

 done:
	free(d);
	return NULL;
}

void fx_call_threads(fxinfo *fx, fx_thread_call call, void *data) {
	fx->cmd_data = data;
	fx->cmd_call = call;
	fx->cmd = FX_CALL;

	// Start threads & wait for them
	LOG("Start cmd");
	pthread_barrier_wait(&fx->cmd_barrier);
	LOG("Wait threads");
	pthread_barrier_wait(&fx->cmd_barrier);
	LOG("Threads done");
}
#endif


// debug stuff

void * __attribute__ ((aligned(64))) load_pam(const char *name, int *width, int *height, int *depth) {
        FILE *fp = fopen(name, "rb");
        char line[32];
        char *p;
        char *mem;

        if (!fp)
                return NULL;

        p = fgets(line, 31, fp);
        if (!p)
                goto closefail;

        if (p[0] == 'P' && p[1] == '5') {
                p = fgets(line, 31, fp);
                if (!p)
                        goto closefail;
                if (sscanf(p, "%d %d", width, height) != 2)
                        goto closefail;

                mem = xmemalign(64, *width * *height);
                if (!mem)
                        goto closefail;
                // ignore bits
                fgets(line, 31, fp);
                fread(mem, 1, (*width)*(*height), fp);
                fclose(fp);

                if (depth)
                        *depth = 1;

                return mem;
        } else if (p[0] == 'P' && p[1] == '6') {
                // colour thing
                p = fgets(line, 31, fp);
                if (!p)
                        goto closefail;
                if (sscanf(p, "%d %d", width, height) != 2)
                        goto closefail;

                mem = xmemalign(64, *width * *height * 3);
                if (!mem)
                        goto closefail;
                // ignore bits
                fgets(line, 31, fp);
                fread(mem, 1, (*width)*(*height)*3, fp);
                fclose(fp);

                if (depth)
                        *depth = 3;

                return mem;
        }

 closefail:
        if (fp)
                fclose(fp);

        return NULL;
}

void save_pam(const char *name, char *mem, int width, int height, int depth) {
        FILE *fp = fopen(name, "wb");

        if (!fp)
                return;

        switch (depth) {
        case 1:
                fprintf(fp, "P5\n%d %d\n255\n", width, height);
                fwrite(mem, 1, width*height, fp);
                break;
        case 3:
                fprintf(fp, "P6\n%d %d\n255\n", width, height);
                fwrite(mem, 3, width*height, fp);
                break;
        }

        fclose(fp);
}

void rgb_to_rgba(uint8_t *src, uint8_t *dst, int iwidth, int iheight) {
	for (int y=0;y<iheight;y++) {
		for (int x=0;x<iwidth;x++) {
			int i = (x+y*iwidth)*4;
			int j = (x+y*iwidth)*3;

			dst[i+0] = src[j+0];
			dst[i+1] = src[j+1];
			dst[i+2] = src[j+2];	
			dst[i+3] = 255;
		}
	}
}

void rgba_to_rgb(uint8_t *src, uint8_t *dst, int iwidth, int iheight) {
	for (int y=0;y<iheight;y++) {
		for (int x=0;x<iwidth;x++) {
			int i = (x+y*iwidth)*4;
			int j = (x+y*iwidth)*3;

			dst[j+0] = src[i+0];
			dst[j+1] = src[i+1];
			dst[j+2] = src[i+2];	
		}
	}
}
