/*
 * Copyright 2013, The University of Waikato.
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 	* Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 	* Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 	* Neither the name of the organization nor the
 *       names of its contributors may be used to endorse or promote products
 * 	  derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package nz.ac.waikato.picfx;

import android.app.Activity;
import android.content.Intent;
import android.graphics.Bitmap;
import android.graphics.Bitmap.CompressFormat;
import android.graphics.Bitmap.Config;
import android.graphics.BitmapFactory;
import android.graphics.BitmapFactory.Options;
import android.net.Uri;
import android.os.Bundle;
import android.os.Environment;
import android.view.LayoutInflater;
import android.view.Menu;
import android.view.MenuItem;
import android.view.View;
import android.view.View.OnClickListener;
import android.widget.FrameLayout;
import android.widget.ImageView;
import android.widget.LinearLayout;
import android.widget.ProgressBar;
import android.widget.ShareActionProvider;
import android.widget.TextView;
import android.widget.Toast;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.logging.Level;
import java.util.logging.Logger;
import nz.ac.waikato.picfx.fx.Effect;
import nz.ac.waikato.picfx.fx.Effect.EffectInfo;
import nz.ac.waikato.picfx.fx.Effect.EffectListener;
import nz.ac.waikato.picfx.fx.EffectEngine;
import nz.ac.waikato.picfx.util.URIUtils;

/**
 * Main window of application.
 * @author notzed
 */
public class EffectActivity extends Activity implements EffectListener, OnPreviewChanged {

	LinearLayout effect_layout;
	FrameLayout effect_controls;
	ProgressBar effect_busy;
	PreviewView effect_view;
	EffectThread effectThread;
	String inputPath;
	ShareActionProvider shareActionProvider;
	Effect efx;

	@Override
	protected void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);
		setContentView(R.layout.effect);

		effect_controls = (FrameLayout) findViewById(R.id.effect_controls);
		effect_busy = (ProgressBar) findViewById(R.id.progress);
		effect_view = (PreviewView) findViewById(R.id.image);
		effect_layout = (LinearLayout) findViewById(R.id.effect_layout);
		LayoutInflater li = getLayoutInflater();

		// Setup effect list
		for (EffectInfo ei : Effect.effects) {
			View icon = li.inflate(R.layout.effect_layout, null);

			icon.setTag(ei);
			effect_layout.addView(icon);

			ImageView iv = (ImageView) icon.findViewById(R.id.icon);
			iv.setImageResource(ei.iconid);

			TextView tv = (TextView) icon.findViewById(R.id.label);
			tv.setText(ei.titleid);

			icon.setOnClickListener(new OnClickListener() {
				public void onClick(View v) {
					setEffect((EffectInfo) v.getTag());
				}
			});
		}

		effect_view.setOnPreviewChanged(this);

		Intent it = getIntent();

		inputPath = URIUtils.uriToPath(this, it.getData());
		effectThread = new EffectThread();
		effectThread.start();

		// Load the initial image view
		effectThread.enqueue(new EffectRequest(CMD_OPEN, inputPath));
	}

	@Override
	protected void onDestroy() {
		super.onDestroy();

		effectThread.shutdown();
	}

	@Override
	public boolean onCreateOptionsMenu(Menu menu) {
		getMenuInflater().inflate(R.menu.effect, menu);

		//MenuItem item = menu.findItem(R.id.share);

		//shareActionProvider = (ShareActionProvider) item.getActionProvider();

		return true;
	}

	public void cmdApply(MenuItem dummy) {
		// a. Apply current effect to full-resolution image
		// b. Save to a file.
		// c. Update the share thing
		if (efx != null) {
			busy(1);
			effectThread.enqueue(new EffectRequest(CMD_APPLY, efx));
		}
	}

	void setEffect(EffectInfo ei) {
		efx = ei.create();

		System.out.println("effect selected " + efx.getClass().getSimpleName());

		// This is a poor simulation of a radio button group
		for (int i = 0; i < effect_layout.getChildCount(); i++) {
			View v = effect_layout.getChildAt(i);
			boolean state = v.getTag() == ei;

			v.setSelected(state);
		}
		View view = efx.createView(this, this);

		effect_controls.removeAllViews();
		effect_controls.addView(view);

		refreshEffect();
	}

	public void effectUpdated() {
		refreshEffect();
	}

	public void previewChanged(float offx, float offy) {
		refreshEffect();
	}

	void busy(int d) {
		effect_busy.setVisibility(d < 0 ? View.INVISIBLE : View.VISIBLE);
	}

	void refreshEffect() {
		if (efx != null) {
			//(1); fast enough busy not needed
			//effect_busy.setVisibility(View.VISIBLE);
			effectThread.enqueue(new EffectRequest(CMD_PREVIEW, efx,
					Math.round(effect_view.getOffsetX()),
					Math.round(effect_view.getOffsetY())));
		}
	}
	final static int CMD_QUIT = 0;
	final static int CMD_OPEN = 1;
	final static int CMD_PREVIEW = 2;
	final static int CMD_APPLY = 3;

	class EffectRequest implements Runnable {

		final int cmd;
		// inputs
		String path;
		Effect effect;
		// Offset for preview view (TODO: matrix?)
		int offx, offy;
		// outputs
		Uri resultUri;
		Bitmap result;

		public EffectRequest(int cmd, String inputPath) {
			this.cmd = cmd;
			this.path = inputPath;
		}

		public EffectRequest(int cmd, Effect effect) {
			this.cmd = cmd;
			this.effect = effect;
		}

		public EffectRequest(int cmd, Effect effect, int offx, int offy) {
			this.cmd = cmd;
			this.effect = effect;
			this.offx = offx;
			this.offy = offy;
		}

		public void run() {
			busy(-1);

			switch (cmd) {
				case CMD_OPEN:
					if (result != null) {
						effect_view.setSource(result);
					} else {
						// Quit?
						Toast.makeText(EffectActivity.this, "Unable to open image file.", 1000).show();
						finish();
					}
					break;
				case CMD_PREVIEW:
					if (result != null) {
						effect_view.setPreview(result);
					} else {
						Toast.makeText(EffectActivity.this, "Error creating preview.", 500).show();
					}
					break;
				case CMD_APPLY:
					if (resultUri != null) {
						Intent it = new Intent(Intent.ACTION_VIEW);

						it.setDataAndType(resultUri, "image/jpeg");

						it = Intent.createChooser(it, "View result");

						System.out.printf("Start activity for image/jpeg uri %s\n", resultUri);

						startActivity(it);
					} else {
						Toast.makeText(EffectActivity.this, "Error creating image.", 1000).show();
					}
					break;
			}
		}
	}

	/**
	 * Thread for creating effect previews.
	 */
	class EffectThread extends Thread {

		LinkedBlockingQueue<EffectRequest> q = new LinkedBlockingQueue<EffectRequest>();
		int previewWidth = 512;
		int previewHeight = 512;
		String inputPath;
		Bitmap inputImage;
		// These are the both same size: previewWidth x previewHeight
		Bitmap previewImage;
		Bitmap resultImage;
		Uri resultUri;
		boolean cancelled = false;
		// Track cached preview state
		Bitmap previewSource;
		int previewX;
		int previewY;
		Effect previewEffect;

		public EffectThread() {
			resultImage = Bitmap.createBitmap(previewWidth, previewHeight, Config.ARGB_8888);
			//previewImage = Bitmap.createBitmap(previewWidth, previewHeight, Config.ARGB_8888);
		}

		public void shutdown() {
			cancelled = true;
			q.clear();
			try {
				interrupt();
				join();
			} catch (InterruptedException ex) {
				Logger.getLogger(EffectActivity.class.getName()).log(Level.SEVERE, null, ex);
			}
		}

		public void enqueue(EffectRequest req) {
			try {
				q.clear();
				q.put(req);
			} catch (InterruptedException ex) {
				Logger.getLogger(EffectActivity.class.getName()).log(Level.SEVERE, null, ex);
			}
		}

		@Override
		public void run() {
			out:
			while (!cancelled) {
				try {
					EffectRequest cmd;
					EffectRequest opencmd = null;
					EffectRequest effectcmd = null;
					EffectRequest applycmd = null;

					cmd = q.take();

					poll:
					while (cmd != null) {

						switch (cmd.cmd) {
							case CMD_OPEN:
								effectcmd = null;
								opencmd = cmd;
								break;
							case CMD_APPLY:
								effectcmd = null;
								applycmd = cmd;
								break poll;
							case CMD_PREVIEW:
								effectcmd = cmd;
								break;
							case CMD_QUIT:
								break out;
						}

						cmd = q.poll();
					}

					// Process open commands first
					if (opencmd != null) {
						try {
							BitmapFactory.Options opts = new Options();

							// Force 32-bit image
							opts.inPreferredConfig = Config.ARGB_8888;

							inputPath = opencmd.path;

							System.out.println("Loading src: " + inputPath);
							inputImage = BitmapFactory.decodeFile(inputPath, opts);

							opencmd.result = inputImage;
						} finally {
							runOnUiThread(opencmd);
						}
					}

					// Process effect command next for better responsiveness
					if (effectcmd != null && inputImage != null) {
						try {
							if (true) {
								// This is about 30% faster when just changing parameters
								EffectEngine ee = EffectEngine.get();
								if (previewSource != inputImage
										|| previewX != effectcmd.offx
										|| previewY != effectcmd.offy
										|| previewEffect != effectcmd.effect) {
									previewSource = inputImage;
									previewX = effectcmd.offx;
									previewY = effectcmd.offy;
									previewEffect = effectcmd.effect;
									//ee.loadPreview(inputImage, previewX, previewY);
									effectcmd.effect.loadPreview(ee, inputImage, previewX, previewY);
								}
								effectcmd.effect.updatePreview(ee, resultImage);
								effectcmd.result = resultImage;
							} else {
								// This should be a shallow copy
								previewImage = Bitmap.createBitmap(inputImage, effectcmd.offx, effectcmd.offy, previewWidth, previewHeight);

								effectcmd.effect.apply(previewImage, resultImage);
								effectcmd.result = resultImage;
							}
						} finally {
							runOnUiThread(effectcmd);
						}
					}

					// Finally process apply commands
					if (applycmd != null && inputImage != null) {
						try {
							Bitmap output = Bitmap.createBitmap(inputImage.getWidth(), inputImage.getHeight(), Config.ARGB_8888);

							// Run it
							applycmd.effect.apply(inputImage, output);

							// Create a unique save file name
							File savedir = Environment.getExternalStoragePublicDirectory(android.os.Environment.DIRECTORY_PICTURES);
							String srcname = new File(inputPath).getName();
							int dot = srcname.lastIndexOf(".");

							savedir.mkdirs();

							if (dot != -1) {
								srcname = srcname.substring(0, dot);
							}
							File dstfile;
							int id = 0;
							do {
								dstfile = new File(savedir, String.format("%s_%03d.jpeg", srcname, id++));
							} while (dstfile.exists());

							System.out.printf("Writing to: %s\n", dstfile);

							// Write it out
							FileOutputStream fos = null;
							try {
								fos = new FileOutputStream(dstfile);
								output.compress(CompressFormat.JPEG, 90, fos);
								applycmd.resultUri = Uri.fromFile(dstfile);
							} catch (IOException ex) {
							} finally {
								if (fos != null) {
									try {
										fos.close();
									} catch (IOException ex1) {
										Logger.getLogger(EffectActivity.class.getName()).log(Level.SEVERE, null, ex1);
									}
								}
							}
						} finally {
							runOnUiThread(applycmd);
						}
					}
				} catch (InterruptedException ex) {
				} catch (Exception ex) {
					// TODO: out of memory stuff?
					// TODO: handle cleanup of busy state
					ex.printStackTrace();
				}
			}
		}
	}
}
