/*
 * Copyright 2013, The University of Waikato.
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 	* Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 	* Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 	* Neither the name of the organization nor the
 *       names of its contributors may be used to endorse or promote products
 * 	  derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package nz.ac.waikato.picfx;

import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.Paint;
import android.util.AttributeSet;
import android.view.MotionEvent;
import android.view.View;

/**
 * Image view that pans an image and allows for a delay-generated image to
 * replace it when ready.
 * @author notzed
 */
public class PreviewView extends View {

	OnPreviewChanged onPreviewChanged;
	int previewWidth = 512;
	int previewHeight = 512;

	public PreviewView(Context context) {
		super(context);
		init();
	}

	public PreviewView(Context context, AttributeSet attrs) {
		super(context, attrs);
		init();
	}

	public PreviewView(Context context, AttributeSet attrs, int defStyle) {
		super(context, attrs, defStyle);
		init();
	}

	private void init() {
	}

	public void setOnPreviewChanged(OnPreviewChanged onPreviewChanged) {
		this.onPreviewChanged = onPreviewChanged;
	}

	public void setSource(Bitmap bm) {
		preview = null;
		source = bm;

		updateTiles(bm);

		invalidate();
	}

	public void setPreview(Bitmap bm) {
		poffx = 0;
		poffy = 0;
		preview = bm;

		previewAlpha = 255;
		invalidate();
	}

	public float getOffsetX() {
		return -offx;
	}

	public float getOffsetY() {
		return -offy;
	}

	void startPanning() {
		// Well, I could pan with it?
		//processedBitmap = null;
		invalidate();
	}
	int previewAlpha = 255;
	Bitmap source;
	Bitmap preview;
	float offx, offy;
	float startoffx, startoffy;
	float startx, starty;
	float poffx, poffy;
	Paint overPaint = new Paint();

	/**
	 * Android can't seem to handle drawing big bitmaps, so one has to do it all manually.
	 */
	static class Tile {

		int x, y;
		Bitmap bm;

		public Tile(int x, int y, Bitmap bm) {
			this.x = x;
			this.y = y;
			this.bm = bm;
		}
	}
	final int tsize = 512;
	int tw, th;
	Tile[] tiles;

	void updateTiles(Bitmap bm) {
		tw = (bm.getWidth() + tsize - 1) / tsize;
		th = (bm.getHeight() + tsize - 1) / tsize;
		tiles = new Tile[tw * th];

		System.out.printf("image size %dx%d tiles %dx%d\n", bm.getWidth(), bm.getHeight(), tw, th);

		for (int y = 0; y < th; y++) {
			for (int x = 0; x < tw; x++) {
				int x0 = x * tsize;
				int y0 = y * tsize;
				int x1 = Math.min(bm.getWidth(), x0 + tsize);
				int y1 = Math.min(bm.getHeight(), y0 + tsize);

				System.out.printf(" create tile %2d,%2d = %d-%d %d-%d\n", x, y, x0, x1, y0, y1);

				tiles[x + y * tw] = new Tile(x0, y0, Bitmap.createBitmap(source, x0, y0, x1 - x0, y1 - y0));
			}
		}
	}

	void drawTiles(Canvas canvas, int x, int y, int w, int h) {
		int tx = x / tsize;
		int ty = y / tsize;

		System.out.printf("draw tiles at %dx%d\n", x, y);

		for (int my = ty * tsize - y; my < h && ty < th; my += tsize, ty += 1) {
			for (int mx = tx * tsize - x, ox = tx; mx < w && ox < tw; mx += tsize, ox += 1) {
				Tile t = tiles[ox + ty * tw];

				canvas.drawBitmap(t.bm, mx, my, null);
			}
		}
	}

	@Override
	protected void onDraw(Canvas canvas) {
		super.onDraw(canvas);

		canvas.save();
		if (true) {
			if (tiles != null) {
				drawTiles(canvas, (int) -offx, (int) -offy, getWidth(), getHeight());
				if (preview != null) {
					//overPaint.setAlpha(previewAlpha);
					canvas.drawBitmap(preview, poffx, poffy, overPaint);
				}
			}
		} else {
			if (source != null) {
				canvas.drawBitmap(source, offx, offy, null);

				int x0 = (int) -offx;
				int y0 = (int) -offy;
				int x1 = Math.min(source.getWidth(), x0 + getWidth());
				int y1 = Math.min(source.getHeight(), y0 + getHeight());

				//Bitmap bm = Bitmap.createBitmap(source, x0, y0, x1-x0, y1-y0);

				//canvas.drawBitmap(bm, 0, 0, null);
				//canvas.drawBitmap(source, new Rect(x0, y0, x1, y1), new Rect(0, 0, x1 - x0, y1 - y0), null);
				if (preview != null) {
					//overPaint.setAlpha(previewAlpha);
					canvas.drawBitmap(preview, poffx, poffy, overPaint);
				}
			}
		}
		canvas.restore();
	}

	void onPreviewChanged(float offx, float offy) {
		// clear whilst re-calculating
		setPreview(null);
		if (onPreviewChanged != null)
			onPreviewChanged.previewChanged(offx, offy);
	}

	@Override
	public boolean onTouchEvent(MotionEvent ev) {
		switch (ev.getActionMasked()) {
			// this is for multi-finger things
			case MotionEvent.ACTION_POINTER_DOWN:
				break;
			case MotionEvent.ACTION_POINTER_UP:
				break;
			case MotionEvent.ACTION_DOWN:
				if (ev.getActionIndex() == 0) {
					startx = ev.getX();
					starty = ev.getY();
					startoffx = offx;
					startoffy = offy;

					startPanning();

					return true;
				}
				break;
			case MotionEvent.ACTION_MOVE:
				if (ev.getActionIndex() == 0) {
					moveTo(ev.getX(), ev.getY());
					return true;
				}
				break;
			case MotionEvent.ACTION_UP:
				if (ev.getActionIndex() == 0) {
					moveTo(ev.getX(), ev.getY());

					onPreviewChanged(-offx, -offy);
					return true;
				}
				break;
		}

		return super.onTouchEvent(ev);
	}

	void moveTo(float tx, float ty) {
		float deltax = tx - startx;
		float deltay = ty - starty;

		// Track input
		float ox = startoffx + deltax;
		float oy = startoffy + deltay;

		// Track preview
		poffx = Math.min(-startoffx, deltax);
		poffy = Math.min(-startoffy, deltay);

		if (source != null) {
			ox = Math.max(ox, previewWidth - source.getWidth());
			oy = Math.max(oy, previewHeight - source.getHeight());
			poffx = Math.max(poffx, (previewWidth - source.getWidth() - startoffx));
			poffy = Math.max(poffy, (previewHeight - source.getHeight() - startoffy));
		}

		ox = Math.min(ox, 0);
		oy = Math.min(oy, 0);

		if (ox != offx || oy != offy) {
			offx = ox;
			offy = oy;
			invalidate();
		}

	}
}
