/*
 * Copyright 2013, The University of Waikato.
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 	* Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 	* Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 	* Neither the name of the organization nor the
 *       names of its contributors may be used to endorse or promote products
 * 	  derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package nz.ac.waikato.picfx.fx;

import android.content.Context;
import android.graphics.Bitmap;
import android.view.View;
import android.widget.TextView;
import nz.ac.waikato.picfx.R;

/**
 * Base class for effects.
 * @author notzed
 */
public abstract class Effect {

	static final int ksize = 512;

	static {
		EffectEngine.init(ksize, ksize);
	}
	public final static EffectInfo[] effects = new EffectInfo[]{
		new EffectInfo(R.string.fx_blur, R.drawable.swan_blur) {
			@Override
			public Effect create() {
				return new GaussianBlur();
			}
		},
		new EffectInfo(R.string.fx_glare, R.drawable.swan_glare) {
			@Override
			public Effect create() {
				return new Glare();
			}
		},
		new EffectInfo(R.string.fx_glaze, R.drawable.swan_glaze) {
			@Override
			public Effect create() {
				return new Glaze();
			}
		},
		new EffectInfo(R.string.fx_soft_focus, R.drawable.swan_soft_focus) {
			@Override
			public Effect create() {
				return new SoftFocus();
			}
		},
		new EffectInfo(R.string.fx_wiener_deconvolution, R.drawable.swan_wiener_deconvolution) {
			@Override
			public Effect create() {
				return new WienerDeconvolution();
			}
		}
	};

	public View createView(Context c, EffectListener el) {
		TextView tv = new TextView(c);

		tv.setText(getClass().getName());

		return tv;
	}

	public void apply(Bitmap input, Bitmap output) {
		// NULL effect, just for debugging
		int w = input.getWidth();
		int h = input.getHeight();
		int[] row = new int[w];

		for (int y = 0; y < h; y++) {
			input.getPixels(row, 0, w, 0, y, w, 1);
			output.setPixels(row, 0, w, 0, y, w, 1);
		}
	}

	public abstract void updatePreview(EffectEngine ee, Bitmap output);

	public void loadPreview(EffectEngine ee, Bitmap input, int x, int y) {
		ee.loadPreview(input, x, y);
	}

	public interface EffectListener {

		public void effectUpdated();
	}

	public abstract static class EffectInfo {

		/**
		 * Title resource id
		 */
		public final int titleid;
		/**
		 * Icon resource id
		 */
		public final int iconid;

		public EffectInfo(int titleid, int iconid) {
			this.titleid = titleid;
			this.iconid = iconid;
		}

		/**
		 * Instantiate a new effect object
		 */
		public abstract Effect create();
	}
}
