/*
 * Copyright 2013, The University of Waikato.
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 	* Redistributions of source code must retain the above copyright
 * 	  notice, this list of conditions and the following disclaimer.
 * 	* Redistributions in binary form must reproduce the above copyright
 * 	  notice, this list of conditions and the following disclaimer in the
 * 	  documentation and/or other materials provided with the distribution.
 * 	* Neither the name of the organization nor the
 *       names of its contributors may be used to endorse or promote products
 * 	  derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package nz.ac.waikato.picfx.fx;

import android.graphics.Bitmap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

/**
 * Some thoughts on performance/memory:
 *
 * For previews, create a fixed-sized data, say 512x512.
 *
 * A change is then just:
 *  update kernel
 *  fft kernel
 *  apply kernel
 *  ifft result
 *
 * i.e. no need to fft image input
 * @author notzed
 */
public class EffectEngine {
	/**
	 * Gaussian blur.
	 * param[0] = sigma
	 * param[1] = blend factor
	 */
	public final static int EFFECT_BLUR = 0;
	/**
	 * Weienr deconvolution
	 * param[0] = sigma
	 * param[1] = k
	 * ?? blend factor?
	 */
	public final static int EFFECT_WIENER = 1;
	/**
	 * Glare effect
	 * param[0] = sigma
	 * param[1] = r
	 * param[2] = g
	 * param[3] = b
	 * param[4] = amount
	 */
	public final static int EFFECT_GLARE = 2;
	/**
	 * Glaze effect
	 * param[0] = sigma
	 * param[1] = ka
	 * param[2] = kd
	 * param[3] = ks
	 * param[4] = shininess
	 * param[5] = Dr
	 * param[6] = Dg
	 * param[7] = Db
	 * param[8] = Sr
	 * param[9] = Sg
	 * param[10] = Sb
	 * param[11] = light direction x
	 * param[12] = y
	 * param[13] = z
	 * param[14] = viewer direction x
	 * param[15] = y
	 * param[16] = z
	 *
	 */
	public final static int EFFECT_GLAZE = 3;

	static {
		try {
			System.loadLibrary("ffts_jni");
			System.loadLibrary("picfx_jni");
		} catch (Exception ex) {
			System.err.println("Unable to load necessary native libraries: " + ex);
			System.exit(1);
		}
	}
	private long p;
	static private EffectEngine e;
	//
	static int nthreads = Runtime.getRuntime().availableProcessors();
	static ExecutorService pool = Executors.newFixedThreadPool(nthreads);

	private EffectEngine(long p) {
		this.p = p;
	}

	public static void init(int w, int h) {
		if (e == null) {
			e = new EffectEngine(create(w, h, Runtime.getRuntime().availableProcessors()));
		}
	}

	public static EffectEngine get() {
		return e;
	}

	public int getThreadCount() {
		return nthreads;
	}

	public ExecutorService getPool() {
		return pool;
	}

	// Native interfaces
	/**
	 * Create an effect handler
	 * @param fw filter width (power of 2)
	 * @param fh filter height (power of 2)
	 * @param number of threads to use.  Ignored at present.
	 * @return handle
	 */
	protected static native long create(int fw, int fh, int nthreads);

	protected static native void free(long p);

	/**
	 * Set the process, use update to set the parameters
	 * @param p
	 * @param effect
	 */
	protected static native void setProcess(long p, int effect, float[] data, float... params);

	/**
	 * Apply effect to whole image
	 * @param p
	 * @param src
	 * @param dst
	 * @param overlap (TODO: remove this, it's in C)
	 */
	protected static native void process(long p, Bitmap src, Bitmap dst, int overlap);

	/**
	 * Update parameters/pre-calculate anything that needs it
	 * @param p
	 * @param data
	 * @param params
	 */
	protected static native void update(long p, float[] data, float... params);

	/**
	 * Set the image used for a preview.
	 * This stores the forward fft of the image.
	 */
	protected static native void loadPreview(long p, Bitmap src, int x, int y);

	/**
	 * Apply effect to stored preview and generate result.
	 */
	protected static native void processPreview(long p, Bitmap dst);

	public void setProcess(int effect, float[] data, float... params) {
		setProcess(p, effect, data, params);
	}

	public void process(Bitmap src, Bitmap dst, int overlap) {
		process(p, src, dst, overlap);
	}

	public void update(float[] data, float... params) {
		update(p, data, params);
	}

	public void loadPreview(Bitmap src, int x, int y) {
		loadPreview(p, src, x, y);
	}

	public void processPreview(Bitmap dst) {
		processPreview(p, dst);
	}
}
